"""
    OptimizationVariable{T} <: AbstractOptVariable

Mutable composite type for holding unconstrained optimization variables.

# Fields:
- `name::Symbol`  # Symbol with the variable name
- `value::T`  # Current value
- `weight::T`  # Weight
- `bounds::Vector{T}`  # Vector holding the upper and lower bounds of the variable
- `diff_type::Symbol`  # Possible options for diffType are :AutoForwardDiff, :AutoReverseDiff, :CenteredDiff, :ForwardDiff
- `finite_diff_step::T`  # Size of the finite difference step, not used if `diffType` is :AutoForwardDiff or :AutoReverseDiff
- `min_finite_diff_step::T`  # Minimum size of the finite difference step size
- `active::Bool`  # Flag indicating whether variable is active for currrent iteration
- `recompute::Bool`  # Flag indicating if equilibrium recomputation is required upon change of value
"""
mutable struct OptimizationVariable{T} <: AbstractOptVariable
  name::Symbol
  value::T
  weight::T
  bounds::Vector{T}
  diff_type::Symbol
  finite_diff_step::T
  min_finite_diff_step::T
  equilibrium_key::UInt
  active::Bool
  recompute::Bool
  OptimizationVariable() = OptimizationVariable{Float64}()
  OptimizationVariable{T}() where {T} = new{T}()
  OptimizationVariable{T}(n,v,w,b,dt,fd,mfd,k,a,r) where {T} = new(n,v,w,b,dt,fd,mfd,k,a,r)
end

function OptimizationVariable(name::Symbol,
                              value=nothing,
                              equilibrium_key=zero(UInt);
                              weight=1.0,
                              bounds = [-Inf, Inf],
                              diff_type=:CenteredDiff,
                              fd_step=0.0,
                              min_fd_step=0.0,
                              active=true,
                              recompute=true,
                             )
  if isnothing(value)
    throw(error("Incorrect variable specification, variable must have an initial value"))
  end
  lower_bound = bounds[1]
  upper_bound = bounds[2]
  finite_diff_step = fd_step
  if isapprox(value, zero(typeof(value)))
    finite_diff_step = isapprox(fd_step, zero(typeof(fd_step))) ? 1e-4 : fd_step
  else
    if isapprox(fd_step, zero(typeof(fd_step)))
      finite_diff_step = 1e-4 > 1e-4*abs(value) ? 1e-4 : 1e-4*abs(value)
    else
      finite_diff_step = fd_step > fd_step * abs(value) ? fd_step : fd_step * abs(value)
    end
  end
  min_finite_diff_step = isapprox(min_fd_step, zero(typeof(min_fd_step))) ? 1e-3 * finite_diff_step : min_fd_step
  promoted_value, promoted_weight, promoted_lb, promoted_ub, promoted_fd_step, promoted_min_fd_step = promote(value, weight, lower_bound, upper_bound, finite_diff_step,min_finite_diff_step)
  T = typeof(promoted_value)
  return OptimizationVariable{T}(name, promoted_value, promoted_weight, [promoted_lb, promoted_ub],
                                 diff_type, promoted_fd_step, promoted_min_fd_step,
                                 equilibrium_key, active, recompute)
end

function OptimizationVariable(name::AbstractString,
                               value=nothing,
                               equilibrium_key=zero(UInt);
                               kwargs...)
  return OptimizationVariable(Symbol(name), value, equilibrium_key; kwargs...)
end

@inline function name(v::OptimizationVariable)
    String(v.name)
end