using Printf
using Optim
using .MPI
"""
    OptComms

Composite type holding the different communicators for MPI-based optimization.

# Fields:
- global_comm::MPI.Comm  # The top-level communicator between all processors. Rank 0 on this communicator is the global head.
- group_comm::MPI.Comm  # Communicator linking all the group heads.  Rank 0 on this communicator is the global head.
- eval_comm::MPI.Comm  # Communicator linking all the processors in a particular group.  Rank 0 on this communicator is the group head.
"""
struct OptComms
  global_comm::MPI.Comm
  group_comm::MPI.Comm
  eval_comm::MPI.Comm
  ranks::NamedTuple{(:global, :group, :eval), Tuple{Int, Int, Int}}
  #targetComm::MPI.Comm
  #targetGroup::Int
  #chunkLength::Int
  #displacement::Int
end

"""
    setup_comms(procs_per_group,global_comm::MPI.Comm)

Split `global_comm` into groups where the communicator splitting attempts to
adhere as close to `procs_per_group` as possible, returning an `OptComms` type.

See also: [`OptComms`](@ref)
"""
function setup_comms(procs_per_group::Int,
                     global_comm::MPI.Comm)
  global_size = MPI.Comm_size(global_comm)
  global_rank = MPI.Comm_rank(global_comm)
  group_div, group_rem = divrem(global_size-1, procs_per_group)
  n_worker_groups = group_rem == 0 ? group_div : group_rem >= div(group_div, 2) ? group_div + 1 : group_div
  @debug "group_div = $group_div, group_rem = $group_rem, n_worker_groups = $n_worker_groups"
  n_procs_group = collect(1:n_worker_groups) .* procs_per_group
  group_id = global_rank != 0 ? !isnothing(findfirst(global_rank .<= n_procs_group)) ? findfirst(global_rank .<= n_procs_group) : n_worker_groups : 0
  #println(global_rank,' ',group_id)
  #groupSize = nGroups > 1 ? group_id < nGroups-1 ? group_div : group_div + group_rem : global_size
  eval_comm = MPI.Comm_split(global_comm, group_id, global_rank)
  eval_rank = MPI.Comm_rank(eval_comm)
  group_comm = MPI.Comm_split(global_comm, eval_rank, global_rank)
  ranks = NamedTuple{(:global, :group, :eval), Tuple{Int, Int, Int}}((global_rank,group_id,eval_rank))
  #procsPerTarget = procs_per_group
  #targetComm = eval_comm
  return OptComms(global_comm, group_comm, eval_comm, ranks)
end

"""
    optimize_mpi(prob::Problem,comms::OptComms)

Run an optimization problem defined by `prob` using MPI with communicators
defined in `comms`.
"""
function optimize_mpi(prob::Problem{T},
                      comms::OptComms;
                     ) where {T}
    global_rank = comms.ranks.global
    if global_rank == 0
        # Define the callback function that requires knowledge of the problem
        function cb_function(x)
            update_history!(x, prob)
        end

        # Extract initial state vector of BitType data
        initial_state_vector = state_vector_values(prob)

        # Set the intial gradient of each target to zero
        for i in 1:length(prob.targets)
            prob.targets[i].gradient = zeros(T, length(prob.state_vector))
        end

        # Augment the arguments of the optimizer options with the callback
        option_args = Dict{Symbol, Any}(:callback => cb_function)
        for f in fieldnames(Optim.Options)
            val = getfield(prob.optimizer.options, f)
            if !isnothing(val)
                option_args[f] = val
            end
        end
        new_options = (; option_args...)

        log_io = open("stellopt_$(prob.output_tag).log", "w+")
        logger = SimpleLogger(log_io, Logging.Info)
        residual = with_logger(logger) do
            Optim.optimize(Optim.only_fg!((F, G, θ) -> objective_gradient!(F, G, θ, prob, comms; logger_io = log_io)),
                            initial_state_vector, prob.optimizer.method, Optim.Options(; new_options...))
        end
        flush(log_io)
        close(log_io)
        write_history_file(prob)
        MPI.bcast(true, 0, comms.global_comm)
        return residual
    else
        worker_loop(prob, comms)
        return nothing
    end
end

function objective_gradient!(F,
                             G,
                             x::AbstractVector,
                             prob::Problem,
                             comms::OptComms;
                             logger_io::Union{Nothing, IOStream} = nothing,
                            )
    # If the loop is in this function there is still work left to do; bcast false
    MPI.bcast(false, 0, comms.global_comm)
    state_vector = MPI.bcast(x ,0, comms.global_comm)
    adjust_state_vector!(prob, state_vector)
    obj, grad = (Inf, nothing)
    if isnothing(F)
        MPI.bcast(false, 0, comms.global_comm)
    else
        MPI.bcast(true, 0, comms.global_comm)
        obj = compute_objective(prob, comms)
        MPI.bcast(obj, 0, comms.global_comm)
        adjust_objective!(prob, obj)
    end
    if isnothing(G)
        MPI.bcast(false, 0, comms.global_comm)
    else
        if !isnan(obj)
            MPI.bcast(true, 0, comms.global_comm)
            grad = compute_gradient(prob, comms)
            copyto!(G, grad)
            adjust_gradient!(prob, grad)
        else
            MPI.bcast(false, 0, comms.global_comm)
            copyto!(G, fill(NaN, length(x)))
        end
    end
    !isnothing(logger_io) ? flush(logger_io) : nothing
    if !isnothing(F)
        return obj
    end
    nothing
end

"""
    worker_loop(prob::Problem,comms::OptComms)

Listens for optimization state vector and calculation termination from the
global head node.  Exits when a finished value is broadcast from the global
head node.
"""
function worker_loop(prob::Problem,
                     comms::OptComms;
                    )
    finished = false

    while !finished
        # Check whether head is broadcasting if optimization is finished
        finished = MPI.bcast(finished, 0, comms.global_comm)
        if !finished
            # Get updated state vector values from head node
            state_vector = MPI.bcast(nothing, 0, comms.global_comm)
            adjust_state_vector!(prob, state_vector)
            # Check whether objective function calculation is required
            objective_needed = MPI.bcast(nothing, 0, comms.global_comm)
            if objective_needed
                compute_objective(prob, comms)
                obj = MPI.bcast(nothing, 0, comms.global_comm)
                adjust_objective!(prob, obj)
            end
            # Check whether gradient calculation is required
            gradient_needed = MPI.bcast(nothing, 0, comms.global_comm)
            if gradient_needed
                compute_gradient(prob, comms)
            end
        end
    end
end

function compute_objective(prob::Problem{T},
                           comms::OptComms;
                          ) where {T}
    n_worker_groups = MPI.Comm_size(comms.group_comm) - 1

    # Compute the initial equilibrium and objective functions, broadcast
    # the equilibrium data to all processors. At this point all ranks
    # have the equilibrium input data from the previous iteration, thus
    # can immediately call the evaluate function
    if comms.ranks.global == 0
        F = zero(T)
        success = MPI.bcast(nothing, 1, comms.global_comm)
        if success
            for i in keys(prob.equilibrium_wrappers)
                eq_data = MPI.bcast(nothing, 1, comms.global_comm)
                update_equilibrium!(prob.equilibrium_wrappers[i], eq_data)
                update_input!(prob.equilibrium_wrappers[i])
                eq_mod = lowercase(string(equilibrium_type(prob.equilibrium_wrappers[i])))
                iteration = isassigned(prob.history, 1) ? prob.history[end].iteration + 1 : 0
                write_input_file(prob.equilibrium_wrappers[i], "input_$(eq_mod).opt_$(iteration)")
            end
            for i in 1:length(prob.targets)
                target = MPI.bcast(nothing, 1, comms.global_comm)
                adjust_target_value!(prob, i, target)
            end
            F = MPI.bcast(nothing, 1, comms.global_comm)
        else
            F = convert(T, NaN)
        end
        return F
    elseif comms.ranks.group == 1
        valid_eqs = fill(true, length(prob.equilibrium_wrappers))
        j = 0
        # Do the calculation on the first worker group
        for i in keys(prob.equilibrium_wrappers)
            j += 1
            translate!(prob.equilibrium_wrappers[i], prob.state_vector)
            eq = compute_equilibrium(prob.equilibrium_wrappers[i], comms.eval_comm)
            if !isnothing(eq) && typeof(eq) <: equilibrium_type(prob.equilibrium_wrappers[i])
                update_equilibrium!(prob.equilibrium_wrappers[i], eq)
                update_input!(prob.equilibrium_wrappers[i])
            else
                valid_eqs[j] = false
            end
        end

        if all(valid_eqs)
            compute_targets!(prob)
            MPI.bcast(true, 1, comms.global_comm)
            for i in keys(prob.equilibrium_wrappers)
                MPI.bcast(prob.equilibrium_wrappers[i].eq, 1, comms.global_comm)
            end
            for i in 1:length(prob.targets)
                MPI.bcast(prob.targets[i].current_value, 1, comms.global_comm)
            end
            if comms.ranks.eval == 0
                compute_targets!(prob)
                F = least_squares_objective(prob)
                MPI.bcast(F, 1, comms.global_comm)
            else
                F = MPI.bcast(nothing, 1, comms.global_comm)
            end
        else
            MPI.bcast(false, 1, comms.global_comm)
        end
        return nothing
    else
        success = MPI.bcast(nothing, 1, comms.global_comm)
        if success
            for i in keys(prob.equilibrium_wrappers)
                eq = MPI.bcast(nothing, 1, comms.global_comm)
                update_equilibrium!(prob.equilibrium_wrappers[i], eq)
                update_input!(prob.equilibrium_wrappers[i])
            end
            for i in 1:length(prob.targets)
                target = MPI.bcast(nothing, 1, comms.global_comm)
                adjust_target_value!(prob, i, target)
            end
            F = MPI.bcast(nothing, 1, comms.global_comm)
        end
        return nothing
    end
end

#=
function compute_targets!(prob::Problem,
                          comms::OptComms;
                          wrapper_key::UInt = zero(UInt),
                         )
    geometries = Dict{UInt, AbstractMagneticEquilibrium}()
    if comms.ranks.eval == 0
        extract_target_geometries!(geometries, prob; wrapper_key = wrapper_key)
        MPI.bcast(geometries, 0, comms.eval_comm)
    else
        geometries = MPI.bcast(nothing, 0, comms.eval_comm)
    end
    compute_targets!(prob, geometries, comms)
end

function compute_targets!(prob::Problem,
                          geometries::Dict{UInt, T},
                          comms::OptComms;
                         ) where {T <: AbstractMagneticEquilibrium}
    if comms.ranks.eval == 0
        for i in 1:length(prob.targets)
            target = prob.targets[i]
            if haskey(geometries, target.geometry_hash)
                @idebug "Target value before: $(target.current_value)"
                target_method = copy(target.method)
                geom_slot = findfirst(x -> x === :geometry, target_method.args)
                target_method.args[geom_slot] = geometries[target.geometry_hash]
                MPI.bcast(target_method, 0, comms.eval_comm)
                MPI.bcast(target.target_mod, 0, comms.eval_comm)
                value = evaluate_target_expression(target.target_mod, target_method, comms.eval_comm)
                adjust_value!(prob, i, value)
                MPI.bcast(i < length(prob.targets) ? false : true, 0, comms.eval_comm)
                @debug"Target value after: $(target.current_value)"
            end
        end
    else
        finished = false
        while !finished
            target_method = MPI.bcast(nothing, 0, comms.eval_comm)
            target_mod = MPI.bcast(nothing, 0, comms.eval_comm)
            evaluate_target_expression(target_mod, target_method, comms.eval_comm)
            finished = MPI.bcast(nothing, 0, comms.eval_comm)
        end
    end
end
=#

"""
    compute_gradient(x::AbstractVector,prob::Problem,comms::OptComms)

The main function for computing the objective function and gradient value using MPI.
The global head node acts as a controller, sending worker groups which gradient index
to evaluate and gathering the results.  Only the global head node returns value of the
objective and gradient to the optimization algorithm.
"""
function compute_gradient(prob::Problem{T},
                          comms::OptComms;
                         ) where {T}
    global_rank = MPI.Comm_rank(comms.global_comm)
    group_rank = MPI.Comm_rank(comms.group_comm)
    eval_rank = MPI.Comm_rank(comms.eval_comm)
    n_worker_groups = MPI.Comm_size(comms.group_comm) - 1
    gradient_indices = global_rank == 0 ? collect(1:length(prob.state_vector)) : nothing
    # This is a potential vector for bugs if the tags collide due to a hard coded base tag
    recv_result_from_worker_tags = global_rank == 0 ? 19293 .+ collect(1:n_worker_groups) : nothing
    recv_targets_from_worker_tags = global_rank == 0 ? 14293 .+ collect(1:n_worker_groups) : nothing
    recv_index_from_head_tag = global_rank != 0 ? eval_rank == 0 ? 18323 + group_rank : nothing : nothing
    send_result_to_head_tag = global_rank != 0 ? eval_rank == 0 ? 19293 + group_rank : nothing : nothing
    send_targets_to_head_tag = global_rank != 0 ? eval_rank == 0 ? 14293 + group_rank : nothing : nothing
    send_index_to_worker_tags = global_rank == 0 ? 18323 .+ collect(1:n_worker_groups) : nothing
    recv_index_from_head_req = global_rank != 0 ? eval_rank == 0 ? MPI.Request() : nothing : nothing
    recv_result_from_worker_reqs = global_rank == 0 ? fill(MPI.Request(), n_worker_groups) : nothing
    index = zero(Int)
    index_buffer = zeros(T, 1)
    result = zero(T)
    result_buffer = zeros(T, 1)
    target_gradients = Vector{T}(undef, length(prob.targets))
    index_map = global_rank == 0 ? Dict{Int, Int}() : nothing
    G = Vector{T}(undef, length(prob.state_vector))

    # Do the gradient evaluation using the new base equilibrium
    # input parameters as the starting point
    # bcast to all processes
    finished_eval = MPI.bcast(false, 0, comms.global_comm)

    if global_rank != 0
        # This is a worker, doing the calculations
        index = 0
        if eval_rank == 0
            global recv_index_from_head_req, finished_eval
            while !finished_eval
                if MPI.isnull(recv_index_from_head_req)
                    # Initiate new request if current request is null
                    # The tag needs to match the sending tag from the global head
                    global recv_index_from_head_req = MPI.Irecv!(index_buffer, 0, recv_index_from_head_tag, comms.group_comm)
                    @debug "Initialize recv_from_head_req"

                else
                    # Valid request exists
                    recv_flag, _ = MPI.Test!(recv_index_from_head_req)
                    if recv_flag
                        global index = convert(Int, index_buffer[1])
                        MPI.bcast(index, 0, comms.eval_comm)
                        #@debug "EvalRank $eval_rank sent $index to workers"
                        if index >= 1
                            # Do the calculation
                            local_prob = deepcopy(prob)
                            for i in 1:length(local_prob.targets)
                                local_prob.targets[i].gradient = zeros(T, length(local_prob.state_vector))
                            end
                            result = evaluate_fd_gradient!(index, local_prob, comms)
                            result_buffer[1] = result
                            foreach(i -> target_gradients[i] = local_prob.targets[i].gradient[index], keys(local_prob.targets))
                            # Send the results back to the global head
                            MPI.Isend(result_buffer, 0, send_result_to_head_tag, comms.group_comm)
                            # Note this is a blocking send
                            MPI.Send(target_gradients, 0, send_targets_to_head_tag, comms.group_comm)
                            # Reset the recvReq to a null Request
                            global recv_index_from_head_req = MPI.Request()
                        else
                            global finished_eval = true
                        end
                        MPI.bcast(finished_eval, 0, comms.eval_comm)
                    end
                end
            end
        else
            while !finished_eval
                logical_buffer = false
                index = MPI.bcast(nothing, 0, comms.eval_comm)
                if index >= 1
                    # Do the calculation
                    # index gets passed here, but instead split this up so that you can Mpi_send prob.equilibrium
                    local_prob = deepcopy(prob)
                    evaluate_fd_gradient!(index, local_prob, comms)
                end
                global finished_eval = MPI.bcast(logical_buffer, 0, comms.eval_comm)
            end
        end
        return nothing
    else
        # This is the global head node
        # Make initial non-blocking sends, initiate the non-blocking receives
        for group_index = 1:n_worker_groups
            @info "Evaluating gradient no. $(group_index) out of $(length(prob.state_vector))"
            grad_index = gradient_indices[group_index]
            index_buffer[1] = grad_index
            index_map[group_index] = gradient_indices[group_index]
            @debug "Sending $(grad_index) to group_index $group_index"
            MPI.Isend(index_buffer, group_index, send_index_to_worker_tags[group_index], comms.group_comm)
            gradient_indices[group_index] = -1
            recv_result_from_worker_reqs[grad_index] = MPI.Irecv!(result_buffer, group_index, recv_result_from_worker_tags[group_index], comms.group_comm)
        end
        # Start listening for receives
        while !finished_eval
            for group_index = 1:n_worker_groups
                # Initiate the first non-blocking recieve of the results
                if !MPI.isnull(recv_result_from_worker_reqs[group_index])
                    # Valid request exists
                    recv_flag, _ = MPI.Test!(recv_result_from_worker_reqs[group_index])
                    if recv_flag
                        index = index_map[group_index]
                        # Assign the gradient result if index > 0
                        result = result_buffer[1]                        
                        G[index] = result
                        @debug "Recieved $(result) from  $group_index"
                        MPI.Recv!(target_gradients, group_index, recv_targets_from_worker_tags[group_index], comms.group_comm)
                        for i in eachindex(target_gradients)
                            prob.targets[i].gradient[index] = target_gradients[i]
                        end
                        # If there are no more gradient points to compute, findfirst will return nothing
                        grad_index = findfirst(x -> x != -1, gradient_indices)
                        @debug gradient_indices
                        if !isnothing(grad_index)
                            @info "Evaluating gradient no. $(grad_index) out of $(length(prob.state_vector))"
                            index_buffer[1] = gradient_indices[grad_index]
                            index_map[group_index] = gradient_indices[grad_index]
                            gradient_indices[grad_index] = -1
                            # Initiate next receive
                            recv_result_from_worker_reqs[group_index] = MPI.Irecv!(result_buffer, group_index, recv_result_from_worker_tags[group_index], comms.group_comm)
                        else
                            # Send the stop command to the workers
                            index_buffer[1] = -1
                            index_map[group_index] = -1
                            # No more work to do
                            recv_result_from_worker_reqs[group_index] = MPI.Request()
                        end
                        @debug "GlobalRank $global_rank sending $(index)  to group_index" group_index
                        MPI.Isend(index_buffer, group_index, send_index_to_worker_tags[group_index], comms.group_comm)
                    end
                end
            end
            if all(MPI.isnull.(recv_result_from_worker_reqs))
                global finished_eval = true
            end
        end
        return G
    end
end

function perturbed_equilibrium(grad_index::Int,
                               perturbation::T,
                               prob::Problem,
                               comms::OptComms;
                               iter_limit::Int=5,
                               backtrack_step=0.5,
                              ) where {T}
    rank = comms.ranks.eval
    initial_value = prob.state_vector[grad_index].value
    wrapper_key = prob.state_vector[grad_index].equilibrium_key
    wrapper = prob.equilibrium_wrappers[wrapper_key]
    iter = 0
    eq = nothing
    perturbed_value = zero(T)
    while isnothing(eq) && iter <= iter_limit
        if rank == 0
            if iter > 0
                @warn "Performing backtracking to recompute equilibrium, consider adjusting step size"
            end
            perturbation *= backtrack_step^iter
            perturbed_value = initial_value + perturbation
            MPI.bcast(perturbed_value, 0, comms.eval_comm)
        else
            perturbed_value = MPI.bcast(nothing, 0, comms.eval_comm)
        end
        adjust_state_vector!(prob, grad_index, perturbed_value)
        translate!(wrapper, prob.state_vector)
        eq = compute_equilibrium(wrapper, comms.eval_comm)
        iter += 1
    end

    if !isnothing(eq)
        update_equilibrium!(wrapper, eq)
        compute_targets!(prob, wrapper_key = wrapper_key) 
        return rank == 0 ? (least_squares_objective(prob), perturbation) : (nothing, perturbation)
    else
        return nothing, perturbation
    end
end

function evaluate_fd_gradient!(grad_index::Int,
                               prob::Problem,
                               comms::OptComms;
                               backtrack_step=0.1,
                              )
    diff_type = prob.state_vector[grad_index].diff_type
    if diff_type === :CenteredDiff
        return centered_diff!(grad_index, prob, comms; backtrack_step=backtrack_step)
    elseif diff_type === :ForwardDiff
        return forward_diff!(grad_index, prob, comms; backtrack_step=backtrack_step)
    else
        error("Method $(opt_var.diff_type) not a valid finite differencing method")
    end
end

function forward_diff!(grad_index::Int,
                       prob::Problem{T},
                       comms::OptComms;
                       backtrack_step=0.1,
                      ) where {T}
    rank = comms.ranks.eval
    opt_var = prob.state_vector[grad_index]
    initial_value = opt_var.value
    perturbation = opt_var.finite_diff_step
    finite_diff_step = copy(opt_var.finite_diff_step)
    min_step_size = copy(opt_var.min_finite_diff_step)
    iter_limit = ceil(Int, (log10(min_step_size) - log10(finite_diff_step)) / log10(backtrack_step))
    target_values = rank == 0 ? [prob.targets[i].current_value for i in 1:length(prob.targets)] : nothing
    if rank == 0
        @debug "Computing at most $iter_limit attempts for derivative"
        @debug "Objective: $(prob.objective)"
    end
    target, perturbation =  perturbed_equilibrium(grad_index, perturbation, prob,
                                                    comms, backtrack_step=backtrack_step)
    opt_var.finite_diff_step = abs(perturbation)
    opt_var.value = initial_value
    if rank == 0 && !isnothing(target)
        for i in eachindex(prob.targets)
            prob.targets[i].gradient[grad_index] = (prob.targets[i].current_value - target_values[i]) / perturbation
        end
        return (target - prob.objective) / finite_diff_step
    else
        return zero(T)
    end
end

function centered_diff!(grad_index::Int,
                        prob::Problem{T},
                        comms::OptComms;
                        backtrack_step=0.1,
                       ) where {T}
    rank = comms.ranks.eval
    opt_var = prob.state_vector[grad_index]
    initial_value = opt_var.value
    compute_backward, compute_forward = (true, true)
    objective_values = Vector{T}(undef, 2)
    target_values = zeros(T, length(prob.targets), 2)
    finite_diff_step = copy(opt_var.finite_diff_step)
    min_step_size = copy(opt_var.min_finite_diff_step)
    if rank == 0
        @debug "Finite difference step: $finite_diff_step"
    end
    while compute_backward || compute_forward
        # Do the negative step
        if compute_backward
        perturbation = -finite_diff_step
        iter_limit = ceil(Int, (log10(min_step_size) - log10(finite_diff_step)) / log10(backtrack_step))
        if rank == 0
            @debug "Computing at most $iter_limit attempts for derivative"
        end
        objective, perturbation =  perturbed_equilibrium(grad_index, perturbation, prob, comms,
                                                        backtrack_step=backtrack_step,
                                                        iter_limit=iter_limit)
        opt_var.value = initial_value
        if rank == 0
            if !isnothing(objective)
            objective_values[1] = objective
            for i in 1:size(target_values, 1)
                target_values[i, 1] = prob.targets[i].current_value
            end
            else
                objective_values[1] = zero(T)
            end
            @debug "$(abs(perturbation)) $(finite_diff_step)"
        end
        if abs(perturbation) != finite_diff_step
            if abs(perturbation) >= min_step_size
                finite_diff_step = copy(abs(perturbation))
                compute_forward = true
                if rank == 0
                    @debug "Finite difference step size adjust to $finite_diff_step"
                end
            else
                finite_diff_step = copy(min_step_size)
                compute_forward = false
                if rank == 0
                    objective_values = zeros(T, 2)
                end
            end
        end
        compute_backward = false
        end

        # Do the positive step
        if compute_forward
        perturbation_vector = finite_diff_step
        iter_limit = ceil(Int, (log10(min_step_size) - log10(finite_diff_step)) / log10(backtrack_step))
        if rank == 0
            @debug "Computing at most $iter_limit attempts for derivative"
        end
        objective, perturbation =  perturbed_equilibrium(grad_index, perturbation, prob, comms,
                                                        backtrack_step=backtrack_step,
                                                        iter_limit=iter_limit)
        if rank == 0
            if !isnothing(objective)
                objective_values[2] = objective
                for i in 1:size(target_values, 1)
                    target_values[i, 2] = prob.targets[i].current_value
                end
            else
                objective_values[2] = zero(typeof(prob.objective))
            end
            @debug "$(abs(perturbation)) $(finite_diff_step)"
        end
        if abs(perturbation) != finite_diff_step 
            if abs(perturbation) >= min_step_size
                finite_diff_step = copy(abs(perturbation))
                compute_backward = true
                if rank == 0
                    @debug "Finite difference step size adjust to $finite_diff_step"
                end
            else
                finite_diff_step = copy(min_step_size)
                compute_backward = false
                if rank == 0
                    objective_values = zeros(T, 2)
                end
            end
        end
        compute_forward = false
        end
    end
    #opt_var.finite_diff_step = finite_diff_step

    opt_var.value = initial_value
    if rank == 0
        for i in 1:size(target_values, 1)
            prob.targets[i].gradient[grad_index] = 0.5 * (target_values[i, 2] - target_values[i, 1]) / finite_diff_step
        end
        return 0.5 * (objective_values[2] - objective_values[1]) / finite_diff_step
    else
        return zero(T)
    end
end

function least_squares_objective(prob::Problem{T}) where {T}
    obj = zero(T)
    for (_, target) in prob.targets
        if isa(target.target_value, Tuple) || isa(target.target_value, AbstractVector)
            for i = 1:length(target.target_value)
                obj += (target.target_value[i] - target.current_value[i])^2 / (target.target_weight[i]^2)
            end
        else
            obj += (target.target_value - target.current_value)^2 / (target.target_weight^2)
        end
    end
    @info "Objective value: $obj"
    return obj
end
