
function optimize_threads(prob::Problem{T}) where {T}
    # Define the callback function that requires knowledge of the problem
    function cb_function(x)
        update_history!(x, prob)
    end

    # Extract initial state vector of BitType data
    initial_state_vector = state_vector_values(prob)

    # Set the intial gradient of each target to zero
    for i in 1:length(prob.targets)
        prob.targets[i].gradient = zeros(T, length(prob.state_vector))
    end

    # Augment the arguments of the optimizer options with the callback
    option_args = Dict{Symbol, Any}(:callback => cb_function)
    for f in fieldnames(Optim.Options)
        val = getfield(prob.optimizer.options, f)
        if !isnothing(val)
            option_args[f] = val
        end
    end
    new_options = (; option_args...)

    log_io = open("stellopt_$(prob.output_tag).log", "w+")
    logger = SimpleLogger(log_io, Logging.Info)
    residual = with_logger(logger) do
         Optim.optimize(Optim.only_fg!((F, G, θ) -> objective_gradient!(F, G, θ, prob; logger_io = log_io)),
                        initial_state_vector, prob.optimizer.method, Optim.Options(; new_options...))
    end
    flush(log_io)
    close(log_io)
    write_history_file(prob)
    return residual
end

function objective_gradient!(F,
                             G,
                             θ::AbstractVector,
                             prob::Problem;
                             logger_io::Union{Nothing, IOStream} = nothing,
                            )
    adjust_state_vector!(prob, θ)
    obj, grad = (Inf, nothing)
    if !isnothing(F)
        obj = compute_objective(prob)
        adjust_objective!(prob, obj)
    end
    if !isnothing(G)
        if !isnan(obj)
            grad = compute_gradient(prob)
            copyto!(G, grad)
            adjust_gradient!(prob, grad)
        else
            copyto!(G, fill(NaN, length(x)))
        end
    end
    !isnothing(logger_io) ? flush(logger_io) : nothing
    if !isnothing(F)
        return obj
    end
    nothing
end

function compute_objective(prob::Problem;
                          )
    valid_eqs = fill(true, length(prob.equilibrium_wrappers))
    j = 0
    for i in eachindex(prob.equilibrium_wrappers)
        j += 1
        translate!(prob.equilibrium_wrappers[i], prob.state_vector)
        eq = compute_equilibrium(prob.equilibrium_wrappers[i])
        if !isnothing(eq)
            update_equilibrium!(prob.equilibrium_wrappers[i], eq)
            eq_mod = lowercase(string(equilibrium_type(prob.equilibrium_wrappers[i])))
            iteration = isassigned(prob.history, 1) ? prob.history[end].iteration + 1 : 0
            write_input_file(prob.equilibrium_wrappers[i], "input_$(eq_mod).opt_$(iteration)")
        else
            valid_eqs[j] = false
        end
    end

    if all(valid_eqs)
        compute_targets!(prob)
        F = least_squares_objective(prob)
        return F
    else
        return NaN
    end
end

function compute_gradient(prob::Problem{T};
                         ) where {T}
    G = zeros(T, length(prob.state_vector))
    for grad_index in eachindex(G)
        local_prob = deepcopy(prob)
        G[grad_index] = evaluate_fd_gradient!(grad_index, local_prob)
        for i in eachindex(prob.targets)
            prob.targets[i].gradient[grad_index] = copy(local_prob.targets[i].gradient[grad_index])
        end
    end
    return G
end

#No backtrack_step input option
function evaluate_fd_gradient!(grad_index::Int,
                               prob::Problem;
                              )
    diff_type = prob.state_vector[grad_index].diff_type
    if diff_type === :CenteredDiff
        return centered_diff!(grad_index, prob)
    elseif diff_type === :ForwardDiff
        return forward_diff!(grad_index, prob)
    else
        error("Method $(diff_type) not a valid finite differencing method")
    end
end

function perturbed_equilibrium(grad_index::Int,
                               perturbation::T,
                               prob::Problem;
                               iter_limit::Int=5,
                               backtrack_step=0.5,
                              ) where{T}
    eq = nothing
    initial_value = prob.state_vector[grad_index].value
    #Just a quick something something. Need to generalized for multiple equilibira
    wrapper_key = prob.state_vector[grad_index].equilibrium_key
    wrapper = prob.equilibrium_wrappers[wrapper_key]
    iter = 0
    while isnothing(eq) && iter <= iter_limit
        if iter > 0
            @warn "Performing backtracking to recompute equilibrium, consider adjusting step size"
        end
        perturbation = backtrack_step^iter .* perturbation
        perturbed_value = initial_value + perturbation
        @debug "grad index: $grad_index, value: $perturbed_value"
        adjust_state_vector!(prob, grad_index, perturbed_value)
        #Just a quick something something. Need to generalized for multiple equilibira
        @debug "new value: $(prob.state_vector[grad_index].value)"
        translate!(wrapper, prob.state_vector)
        
        eq = compute_equilibrium(wrapper)
        iter += 1
    end

    if !isnothing(eq)
        update_equilibrium!(wrapper, eq)
        compute_targets!(prob, wrapper_key = wrapper_key, )
        return least_squares_objective(prob), perturbation
    else
        return nothing, perturbation
    end
end

function forward_diff!(grad_index::Int,
                       prob::Problem{T};
                       backtrack_step=0.1,
                      ) where {T}
    opt_var = prob.state_vector[grad_index]
    initial_value = opt_var.value
    perturbation = opt_var.finite_diff_step
    finite_diff_step = copy(opt_var.finite_diff_step)
    min_step_size = copy(opt_var.min_finite_diff_step)
    iter_limit = ceil(Int, (log10(min_step_size) - log10(finite_diff_step)) / log10(backtrack_step))
    target_values = [prob.targets[i].current_value for i in 1:length(prob.targets)]
    @debug "Computing at most $iter_limit attempts for derivative"
    @debug "Objective: $(prob.objective)"
    objective_value, perturbation = perturbed_equilibrium(grad_index, perturbation, prob, 
                                                          backtrack_step=backtrack_step)
    opt_var.finite_diff_step = abs(perturbation)
    opt_var.value = initial_value
    if !isnothing(objective_value)
        for i in eachindex(prob.targets)
            prob.target[i].gradient[grad_index] = (prob.targets[i].current_value - target_values[i]) / perturbation
        end
        return (objective_value - prob.objective) / opt_var.finite_diff_step
    else
        return zero(T)
    end
end

function centered_diff!(grad_index::Int,
                        prob::Problem{T};
                        backtrack_step=0.1,
                       ) where {T}
    opt_var = prob.state_vector[grad_index]
    initial_value = opt_var.value
    compute_backward, compute_forward = (true, true)
    # TODO change the target values to match the types of actual targets
    objective_values = Vector{T}(undef, 2)
    target_values = zeros(T, length(prob.targets), 2)
    finite_diff_step = copy(opt_var.finite_diff_step)
    min_step_size = copy(opt_var.min_finite_diff_step)
    @debug "Finite difference step: $finite_diff_step"
    while compute_backward || compute_forward
        # Do the negative step
        if compute_backward
            perturbation = -finite_diff_step
            iter_limit = ceil(Int, (log10(min_step_size) - log10(finite_diff_step)) / log10(backtrack_step))
            @debug "Computing at most $iter_limit attempts for derivative"
            objective, perturbation = perturbed_equilibrium(grad_index, perturbation, prob,
                                                            backtrack_step=backtrack_step,
                                                            iter_limit=iter_limit)
            opt_var.value = initial_value
            objective_values[1] = !isnothing(objective) ? objective : zero(typeof(prob.objective))
            if !isnothing(objective)
                objective_values[1] = objective
                for i in 1:size(target_values, 1)
                    target_values[i, 1] = prob.targets[i].current_value
                end
            else
                objective_values[1] = zero(T)
            end
            @debug "$(abs(perturbation)) $(finite_diff_step)"
            if abs(perturbation) != finite_diff_step
                if abs(perturbation) >= min_step_size
                    finite_diff_step = copy(abs(perturbation))
                    compute_forward = true
                    @debug "Finite difference step size adjust to $finite_diff_step"
                else
                    finite_diff_step = copy(min_step_size)
                    compute_forward = false
                    objective_values = zeros(T, 2)
                end
            end
           compute_backward = false
        end

        # Do the positive step
        if compute_forward
            perturbation = finite_diff_step
            iter_limit = ceil(Int, (log10(min_step_size) - log10(finite_diff_step)) / log10(backtrack_step))
            @debug "Computing at most $iter_limit attempts for derivative"
            objective, perturbation =  perturbed_equilibrium(grad_index, perturbation, prob,
                                                             backtrack_step=backtrack_step,
                                                             iter_limit=iter_limit)
            opt_var.value = initial_value
            if !isnothing(objective)
                objective_values[2] = objective
                for i in 1:size(target_values, 1)
                    target_values[i, 2] = prob.targets[i].current_value
                end
            else
                objective_values[2] = zero(typeof(prob.objective))
            end
            @debug "$(abs(perturbation)) $(finite_diff_step)"
            if abs(perturbation) != finite_diff_step 
                if abs(perturbation) >= min_step_size
                    finite_diff_step = copy(abs(perturbation))
                    compute_backward = true
                    @debug "Finite difference step size adjust to $finite_diff_step"
                else
                    finite_diff_step = copy(min_step_size)
                    compute_backward = false
                    objective_values = zeros(T, 2)
                end
            end
            @debug "Target values: $(target_values)"
            compute_forward = false
        end
    end

    @debug "finite_diff_step after backtracking: $(opt_var.finite_diff_step)"
    
    opt_var.value = initial_value
    for i in 1:size(target_values, 1)
        prob.targets[i].gradient[grad_index] = 0.5 * (target_values[i, 2] - target_values[i, 1]) / finite_diff_step
        @debug "$(target_values[i, 2]), $(target_values[i, 1]); $(prob.targets[i].gradient[grad_index]), $(prob.targets[i].gradient)"
    end
    return 0.5 * (objective_values[2] - objective_values[1]) / finite_diff_step
end

function least_squares_objective(prob::Problem{T};
                                 gradient_index::Union{Nothing, Int} = nothing,
                                 finite_diff_step::Union{Nothing, Int} = nothing
                                ) where {T}
    obj = zero(T)
    for (index, target) in prob.targets
        #length(target.label) < 25 ? target_label = target.label * repeat(" ", 25-length(target.label)) : target_label = target.label
        #target_string = @sprintf "%.3E" target.target_value
        #value_string = @sprintf "%.5E" target.current_value
        #sigma_string = @sprintf "%.3E" target.target_weight
        #println(target_label,": target ",target_string," value ",value_string," sigma ",sigma_string)
        if isa(target.target_value, Tuple) || isa(target.target_value, AbstractVector)
            for i = 1:length(target.target_value)
                obj += (target.target_value[i] - target.current_value[i])^2 / (target.target_weight[i]^2)
            end
        else
            obj += (target.target_value - target.current_value)^2 / (target.target_weight^2)
        end
    end
    objective_string = @sprintf "%.5E" obj
    #println("Objective value: ",objective_string)
    return obj
end
