
function write_status_file(prob::Problem)
    status_file = prob.output_directory*prob.tag*"_status.txt"
    if prob.objective_counter == 0
      fid = open(status_file,"w")
      write(fid, "#iteration;target_name;target;value;sigma;(v/s)^2\n")
    else
      fid = open(status_file,"a")
    end
    istring = @sprintf "%3i" prob.objective_counter
    for (index, target) in prob.targets
      @info "Target current value: $(target.current_value)"
      length(target.label) < 25 ? target_label = target.label * repeat(" ", 25-length(target.label)) : target_label = target.label
      target_string = @sprintf "%.7E" target.target_value
      value_string = @sprintf "%.7E" target.current_value
      sigma_string = @sprintf "%.7E" target.target_weight
      vos_string = @sprintf "%.7E" abs(target.current_value - target.target_value)^2/target.target_weight^2
      write(fid,istring*"; "*target_label*"; ")
      write(fid,target_string*"; "*value_string*"; "*sigma_string*"; "*vos_string*"\n")
    end
    close(fid)
end
  
function write_gradients_file(prob::Problem,
                                gradient::AbstractVector;
                               )
    p_iter = prob.gradient_counter
    gradient_file = prob.output_directory*prob.tag*"_gradients.txt"
    if p_iter <= 1
        return
    end
    if p_iter == 2
        fid = open(gradient_file,"w")
        write(fid, "#iteration;varIndex;old_value;new_value;gradient;alpha*P;step\n")
    else
        fid = open(gradient_file,"a")
    end
    nvals = length(prob.state_vector)
    istring = @sprintf "%3i" p_iter
    for j in 1:nvals
        jstring = @sprintf "%3i" j
        nv_string = @sprintf "%.7E" prob.state_vector[j].value
        old_value = prob.history[p_iter-1].state_vector[j]
        ov_string = @sprintf "%.7E" old_value
        grad_string = @sprintf "%.7E" gradient[j]
        alpha_p = (prob.state_vector[j].value - old_value)/gradient[j]
        alpha_string = @sprintf "%.7E" alpha_p
        step = gradient[j] * alpha_p
        sigma_string = @sprintf "%.7E" step
  
        write(fid,istring*"; "*jstring*"; ")
        write(fid,ov_string*"; "*nv_string*"; "*grad_string*"; "*alpha_string*"; "*sigma_string*"\n")
    end
    close(fid)
end

function open_history_file(prob::Problem; )
    name_stem = "stellopt_history_$(prob.output_tag)"
    dir = prob.output_directory * (last(prob.output_directory) != '/' ? "/" : "")
    filename = dir*name_stem*".h5"
    fid = h5open(filename, "w")
    
    attributes(fid)["degrees of freedom"] = length(prob.state_vector)
    attributes(fid)["number of targets"] = length(prob.targets)
    attributes(fid)["jacobian size"] = [length(prob.targets), length(prob.state_vector)]

    attributes(fid)["state vector ids"] = map(v -> name(v), prob.state_vector)
    attributes(fid)["target ids"] = [name(target) for (i, target) in prob.targets]
    fid
end

function write_history_record(fid::HDF5.File,
                              record::OptHistory,
                              count::Int;
                             )
    iter = record.iteration
    create_group(fid, "$count")
    gid = fid["$count"]
    
    id = "Iteration"
    write(gid, id, iter)

    id = "State vector"
    write(gid, id, record.state_vector)

    id = "Total objective"
    write(gid, id, record.objective)

    id = "Gradient"
    write(gid, id, record.gradient)

    id = "Objective terms"
    write(gid, id, record.objective_terms)

    id = "Jacobian"
    write(gid, id, record.jacobian)
    close(gid)
    return nothing
end

function write_history_file(prob::Problem;)
    fid = open_history_file(prob)

    for (i, record) in enumerate(prob.history)
        write_history_record(fid, record, i)
    end
   close(fid)
end