function adjust_time!(prob::Problem, 
                      i::Int,
                      time::Float64;
                    )
    prob.targets[i].runtime = copy(time)
end

# Helper function for adjusting problem value
function adjust_target_value!(prob::Problem, 
                              i::Int,
                              value;
                             )
    prob.targets[i].current_value = copy(value)
end

function adjust_state_vector!(prob::Problem,
                              new_values::Vector;
                             )
    for var_index in eachindex(prob.state_vector)
        set_variable_value!(prob, var_index, new_values[var_index])
    end
    return nothing
end

function adjust_state_vector!(prob::Problem,
                              index::Int,
                              value;
                             )
    prob.state_vector[index].value = value
    return nothing
end

function adjust_objective!(prob::Problem,
                           objective::T;
                          ) where {T}
    prob.objective = copy(objective)
end

function adjust_gradient!(prob::Problem,
                          gradient::Vector{T};
                         ) where {T}
    prob.gradient = copy(gradient)
end

function update_history!(prob::Problem,
                         new_record::OptHistory
                        )
    push!(prob.history, new_record)
end

function update_history!(state,
                         prob::Problem{T};
                        ) where {T}
    n_targets = length(prob.targets)
    n_dofs = length(prob.state_vector)
    jacobian = Matrix{T}(undef, n_dofs, n_targets)

    for t in 1:n_targets
        jacobian[:, t] = copy(prob.targets[t].gradient)
    end

    jacobian = permutedims(jacobian)
    state_vec_values = state_vector_values(prob)
    grad = copy(prob.gradient)
    target_values = [target.current_value for (i, target) in prob.targets]
    record = OptHistory(state.iteration, state_vec_values,
                        prob.objective, grad, target_values, jacobian)
    update_history!(prob, record)
    return false
end

function history(prob::Problem,
                 iter::Int;
                )
    return prob.history[iter]
end

function last_state_vector(prob::Problem)
    return prob.history[end].state_vector
end

function last_gradients(prob::Problem)
    return prob.history[end].gradients
end

function add_target!(prob::Problem,
                     target::AbstractOptTarget;
                    )
    next_index = prob.targets.count + 1
    prob.targets[next_index] = target
    return next_index
end

function add_variable!(prob::Problem,
                       var::AbstractOptVariable;
                      )
    push!(prob.state_vector, var)
    return length(prob.state_vector)
end


function add_variable!(::Problem{T},
                       ::EquilibriumWrapper{T, E},
                       ::Union{AbstractString, Symbol};
                       kwargs...
                      ) where {T, E <: AbstractMagneticEquilibrium}
    error("Method for adding state vector variable from equilibrium $(E) no implemented")
end

function state_vector_values(prob::Problem{T}) where {T}
    values = Vector{T}(undef, length(prob.state_vector))
    for i in eachindex(values, prob.state_vector)
        values[i] = copy(prob.state_vector[i].value)
    end
    return values
end

function add_optimizer!(prob::Problem,
                        optimizer::Optimizer
                       )
    setfield!(prob, :optimizer, optimizer)
    return nothing
end

function set_variable_value!(prob::Problem,
           var_index::Int,
           value::T;
          ) where {T}
    setfield!(prob.state_vector[var_index], :value, value)
    return nothing
end

function add_equilibrium_wrapper!(prob::Problem,
                                  wrapped_eq::EquilibriumWrapper{T, E};
                                 ) where {T, E <: AbstractMagneticEquilibrium}
    key = objectid(wrapped_eq)
    if !haskey(prob.equilibrium_wrappers, key)
        prob.equilibrium_wrappers[key] = wrapped_eq
    end
    return key
end

function update_equilibrium_wrapper!(prob::Problem,
                                     key::UInt,
                                     new_wrapper::EquilibriumWrapper{T, E};
                                    ) where {T, E <: AbstractMagneticEquilibrium}
  prob.equilibrium_wrappers[key] = new_wrapper
  return nothing
end

function add_derived_geometry!(prob::Problem,
                               derived_key::UInt,
                               derived_wrapper::DerivedGeometryWrapper;
                              )
    prob.derived_geometries[derived_key] = derived_wrapper
    add_derived_geometry!(prob.equilibrium_wrappers[derived_wrapper.parent_key],
                          derived_key, derived_wrapper.geometry_expr)
    return derived_key
end

function evaluate_target_expression(::Union{T, Type{T}},
                                    ::Expr;
                                   ) where {T <: AbstractModuleWrapper}
    error("evaluate_target_expression function not implemented in module $T")
end

function extract_target_geometries!(geometries::Dict{UInt, T},
                                    prob::Problem;
                                    wrapper_key::UInt = zero(UInt),
                                   ) where {T <: AbstractMagneticEquilibrium}
    for (key, geom_wrapper) in prob.derived_geometries
        parent_key = geom_wrapper.parent_key
        if wrapper_key == parent_key || iszero(wrapper_key)
            eq_wrapper = prob.equilibrium_wrappers[geom_wrapper.parent_key]
            geometries[key] = extract_derived_geometry(eq_wrapper, key)
        end
    end
end

function compute_targets!(prob::Problem,
                          geometries::Dict{UInt, T};
                         ) where {T <: AbstractMagneticEquilibrium}
    for i in eachindex(prob.targets)
        target = prob.targets[i]
        if haskey(geometries, target.geometry_hash)
            target_method = copy(target.method)
            geom_slot = findfirst(x -> x === :geometry, target_method.args)
            target_method.args[geom_slot] = geometries[target.geometry_hash]
            value = evaluate_target_expression(target.target_mod, target_method)
            adjust_target_value!(prob, i, value)
            #adjust_time!(prob, i, run_time)
        end
    end
end

function compute_targets!(prob::Problem;
                          wrapper_key::UInt = zero(UInt),
                         )
    geometries = Dict{UInt, AbstractMagneticEquilibrium}()
    extract_target_geometries!(geometries, prob; wrapper_key = wrapper_key)
    compute_targets!(prob, geometries)
end

function set_history_file!(prob::Problem,
                           file::HDF5.File;
                          )
    setfield!(prob, :history_file, file)
    return nothing
end